# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/8/2

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The test cases of the accounting application.

"""
from django.test import TestCase

from .forms import TransactionForm


class SortTransactionPostTestCase(TestCase):
    """Tests the sort_post_txn_records() utility."""

    def test_sort(self):
        """Tests the sort_post_txn_records() utility."""
        post = {
            "date": "2020-07-15",
            "notes": "",
            "debit-2-account": "4144",
            "debit-2-ord": "4",
            "debit-2-summary": "",
            "debit-2-amount": "262",
            "debit-3-id": "714703431",
            "debit-3-account": "2715",
            "debit-3-ord": "4",
            "debit-3-summary": "lunch",
            "debit-3-amount": "477",
            "debit-16-id": "541574719",
            "debit-16-account": "6634",
            "debit-16-ord": "2",
            "debit-16-summary": "dinner",
            "debit-16-amount": "525",
            "credit-7-id": "747725334",
            "credit-7-account": "1211",
            "credit-7-ord": "3",
            "credit-7-summary": "",
            "credit-7-amount": "667",
        }
        TransactionForm._sort_post_txn_records(post)
        self.assertEqual(post.get("date"), "2020-07-15")
        self.assertEqual(post.get("notes"), "")
        self.assertEqual(post.get("debit-1-ord"), "1")
        self.assertEqual(post.get("debit-1-id"), "541574719")
        self.assertEqual(post.get("debit-1-account"), "6634")
        self.assertEqual(post.get("debit-1-summary"), "dinner")
        self.assertEqual(post.get("debit-1-amount"), "525")
        self.assertEqual(post.get("debit-2-ord"), "2")
        self.assertEqual(post.get("debit-2-account"), "4144")
        self.assertEqual(post.get("debit-2-summary"), "")
        self.assertEqual(post.get("debit-2-amount"), "262")
        self.assertEqual(post.get("debit-3-ord"), "3")
        self.assertEqual(post.get("debit-3-id"), "714703431")
        self.assertEqual(post.get("debit-3-account"), "2715")
        self.assertEqual(post.get("debit-3-summary"), "lunch")
        self.assertEqual(post.get("debit-3-amount"), "477")
        self.assertEqual(post.get("credit-1-ord"), "1")
        self.assertEqual(post.get("credit-1-id"), "747725334")
        self.assertEqual(post.get("credit-1-account"), "1211")
        self.assertEqual(post.get("credit-1-summary"), "")
        self.assertEqual(post.get("credit-1-amount"), "667")



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/24

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The session-based POST data storage management of the Mia core application.

"""
import random
from typing import Dict, Mapping, Any, Optional

from django.http import HttpResponseRedirect, HttpRequest
from django.shortcuts import redirect

from .utils import UrlBuilder

STORAGE_KEY: str = "stored_post"


def error_redirect(request: HttpRequest, url: str,
                   post: Dict[str, str]) -> HttpResponseRedirect:
    """Redirects to a specific URL on error, with the POST data ID appended
    as the query parameter "s".  The POST data can be loaded with the
    get_previous_post() utility.

    Args:
        request (HttpRequest): The request.
        url (str): The destination URL.
        post (dict[str]): The POST data.

    Returns:
        HttpResponseRedirect: The redirect response.
    """
    post_id = _store(request, post)
    return redirect(str(UrlBuilder(url).query(s=post_id)))


def get_previous_post(request: HttpRequest) -> Optional[Dict[str, str]]:
    """Retrieves the previously-stored POST data.

    Args:
        request (HttpRequest): The request.

    Returns:
        dict: The previously-stored POST data.
    """
    if "s" not in request.GET:
        return None
    return _retrieve(request, request.GET["s"])


def _store(request: HttpRequest, post: Dict[str, str]) -> str:
    """Stores the POST data into the session, and returns the POST data ID that
    can be used to retrieve it later with _retrieve().

    Args:
        request: The request.
        post: The POST data.

    Returns:
        The POST data ID
    """
    if STORAGE_KEY not in request.session:
        request.session[STORAGE_KEY] = {}
    post_id = _new_post_id(request.session[STORAGE_KEY])
    request.session[STORAGE_KEY][post_id] = post
    return post_id


def _retrieve(request: HttpRequest, post_id: str) -> Optional[Dict[str, str]]:
    """Retrieves the POST data from the storage.

    Args:
        request: The request.
        post_id: The POST data ID.

    Returns:
        The POST data, or None if the corresponding data does not exist.
    """
    if STORAGE_KEY not in request.session:
        return None
    if post_id not in request.session[STORAGE_KEY]:
        return None
    return request.session[STORAGE_KEY][post_id]


def _new_post_id(post_store: Mapping[int, Any]) -> str:
    """Generates and returns a new POST ID that does not exist yet.

    Args:
        post_store (dict): The POST storage.

    Returns:
        str: The newly-generated POST ID.
    """
    while True:
        post_id = ""
        while len(post_id) < 16:
            n = random.randint(1, 64)
            if n < 26:
                post_id = post_id + chr(ord("a") + n)
            elif n < 52:
                post_id = post_id + chr(ord("a") + (n - 26))
            elif n < 62:
                post_id = post_id + chr(ord("0") + (n - 52))
            else:
                post_id = post_id + "-_."[n - 62]
        if post_id not in post_store:
            return post_id



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/13

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The template tags and filters of the accounting application.

"""
import re
from decimal import Decimal
from typing import Union, Optional

from django import template
from django.template import RequestContext

from accounting.models import Account
from accounting.utils import ReportUrl
from mia_core.period import Period

register = template.Library()


def _format_positive_amount(value: Union[str, Decimal]) -> str:
    """Formats a positive amount, groups every 3 digits by commas.

    Args:
        value: The amount.

    Returns:
        ReportUrl: The formatted amount.
    """
    s = str(value)
    while True:
        m = re.match("^([1-9][0-9]*)([0-9]{3}.*)", s)
        if m is None:
            break
        s = m.group(1) + "," + m.group(2)
    s = re.sub(r"^(.*\.[0-9]*?)0+$", r"\1", s)
    s = re.sub(r"^(.*)\.$", r"\1", s)
    return s


@register.filter
def accounting_amount(value: Union[Decimal]) -> str:
    """Formats an amount with the accounting notation, grouping every 3 digits
    by commas, and marking negative numbers with brackets instead of signs.

    Args:
        value: The amount.

    Returns:
        ReportUrl: The formatted amount.
    """
    if value is None:
        return ""
    if value == 0:
        return "-"
    s = _format_positive_amount(abs(value))
    if value < 0:
        s = F"({s})"
    return s


@register.filter
def short_amount(value: Union[Decimal]) -> str:
    """Formats an amount, groups every 3 digits by commas.

    Args:
        value: The amount.

    Returns:
        ReportUrl: The formatted amount.
    """
    if value is None:
        return ""
    if value == 0:
        return "-"
    s = _format_positive_amount(abs(value))
    if value < 0:
        s = "-" + s
    return s


@register.simple_tag(takes_context=True)
def report_url(context: RequestContext,
               cash_account: Optional[Account],
               ledger_account: Optional[Account],
               period: Optional[Period]) -> ReportUrl:
    """Returns accounting report URL helper.

    Args:
        context: The request context.
        cash_account: The current cash account.
        ledger_account: The current ledger account.
        period: The period.

    Returns:
        ReportUrl: The accounting report URL helper.
    """
    return ReportUrl(
        namespace=context.request.resolver_match.namespace,
        cash=cash_account or None,
        ledger=ledger_account or None,
        period=period or None)



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/6/30

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The view controllers of the accounting application.

"""
import datetime
import json
import re
from decimal import Decimal
from typing import Dict, Optional, List

from django.conf import settings
from django.contrib import messages
from django.db import transaction
from django.db.models import Sum, Case, When, F, Q, Count, BooleanField, \
    ExpressionWrapper, Exists, OuterRef, Value, CharField
from django.db.models.functions import TruncMonth, Coalesce, Left, StrIndex
from django.http import JsonResponse, HttpResponseRedirect, Http404, \
    HttpRequest, HttpResponse
from django.shortcuts import render, redirect
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _, gettext_noop, gettext
from django.views.decorators.http import require_GET, require_POST
from django.views.generic import ListView, DetailView, TemplateView

from mia_core.period import Period
from mia_core.utils import Pagination, PaginationException, add_default_libs, \
    parse_date
from mia_core.views import DeleteView, FormView, RedirectView
from . import utils
from .forms import AccountForm, TransactionForm, TransactionSortForm
from .models import Record, Transaction, Account

add_default_libs("bootstrap4", "font-awesome-5", "i18n")


@method_decorator(require_GET, name="dispatch")
class CashDefaultView(RedirectView):
    """The default cash account."""
    query_string = True
    pattern_name = "accounting:cash"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["account"] = utils.get_default_cash_account()
        kwargs["period"] = Period.default_spec()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def cash(request: HttpRequest, account: Account,
         period: Period) -> HttpResponse:
    """The cash account.

    Args:
        request: The request.
        account: The account.
        period: The period.

    Returns:
        The response.
    """
    # The accounting records
    if account.code == "0":
        records = list(
            Record.objects
            .filter(
                Q(transaction__in=Transaction.objects.filter(
                    Q(date__gte=period.start),
                    Q(date__lte=period.end),
                    (Q(record__account__code__startswith="11") |
                     Q(record__account__code__startswith="12") |
                     Q(record__account__code__startswith="21") |
                     Q(record__account__code__startswith="22")))),
                ~Q(account__code__startswith="11"),
                ~Q(account__code__startswith="12"),
                ~Q(account__code__startswith="21"),
                ~Q(account__code__startswith="22"))
            .order_by("transaction__date", "transaction__ord",
                      "is_credit", "ord"))
        balance_before = Record.objects \
            .filter(
                Q(transaction__date__lt=period.start),
                (Q(account__code__startswith="11") |
                 Q(account__code__startswith="12") |
                 Q(account__code__startswith="21") |
                 Q(account__code__startswith="22"))) \
            .aggregate(
                balance=Coalesce(Sum(Case(
                    When(is_credit=True, then=-1),
                    default=1) * F("amount")), 0))["balance"]
    else:
        records = list(
            Record.objects
            .filter(
                Q(transaction__in=Transaction.objects.filter(
                    Q(date__gte=period.start),
                    Q(date__lte=period.end),
                    Q(record__account__code__startswith=account.code))),
                ~Q(account__code__startswith=account.code))
            .order_by("transaction__date", "transaction__ord",
                      "is_credit", "ord"))
        balance_before = Record.objects \
            .filter(
                transaction__date__lt=period.start,
                account__code__startswith=account.code) \
            .aggregate(
                balance=Coalesce(Sum(Case(When(
                    is_credit=True, then=-1),
                    default=1) * F("amount")), 0))["balance"]
    balance = balance_before
    for record in records:
        sign = 1 if record.is_credit else -1
        balance = balance + sign * record.amount
        record.balance = balance
    record_sum = Record(
        transaction=(Transaction(date=records[-1].transaction.date)
                     if len(records) > 0
                     else Transaction(date=timezone.localdate())),
        account=account,
        summary=_("Total"),
    )
    record_sum.balance = balance
    record_sum.credit_amount = sum([
        x.amount for x in records if x.is_credit])
    record_sum.debit_amount = sum([
        x.amount for x in records if not x.is_credit])
    record_balance_before = Record(
        transaction=Transaction(date=period.start),
        account=Account.objects.get(code=Account.ACCUMULATED_BALANCE),
        is_credit=balance_before >= 0,
        amount=abs(balance_before),
    )
    record_balance_before.balance = balance_before
    records.insert(0, record_balance_before)
    records.append(record_sum)
    try:
        pagination = Pagination(request, records, True)
    except PaginationException as e:
        return redirect(e.url)
    records = pagination.items
    utils.find_imbalanced(records)
    utils.find_order_holes(records)
    accounts = utils.get_cash_accounts()
    shortcut_accounts = utils.get_cash_shortcut_accounts()
    return render(request, "accounting/report-cash.html", {
        "record_list": records,
        "pagination": pagination,
        "shortcut_accounts": [x for x in accounts
                              if x.code in shortcut_accounts],
        "all_accounts": [x for x in accounts
                         if x.code not in shortcut_accounts],
    })


@method_decorator(require_GET, name="dispatch")
class CashSummaryDefaultView(RedirectView):
    """The default cash account summary."""
    query_string = True
    pattern_name = "accounting:cash-summary"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["account"] = utils.get_default_cash_account()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def cash_summary(request: HttpRequest, account: Account) -> HttpResponse:
    """The cash account summary.

    Args:
        request: The request.
        account: The account.

    Returns:
        The response.
    """
    # The account
    accounts = utils.get_cash_accounts()
    # The month summaries
    if account.code == "0":
        months = [utils.MonthlySummary(**x) for x in Record.objects
                  .filter(
                    Q(transaction__in=Transaction.objects.filter(
                        Q(record__account__code__startswith="11") |
                        Q(record__account__code__startswith="12") |
                        Q(record__account__code__startswith="21") |
                        Q(record__account__code__startswith="22"))),
                    ~Q(account__code__startswith="11"),
                    ~Q(account__code__startswith="12"),
                    ~Q(account__code__startswith="21"),
                    ~Q(account__code__startswith="22"))
                  .annotate(month=TruncMonth("transaction__date"))
                  .values("month")
                  .order_by("month")
                  .annotate(
                    debit=Coalesce(
                        Sum(Case(When(is_credit=False, then=F("amount")))),
                        0),
                    credit=Coalesce(
                        Sum(Case(When(is_credit=True, then=F("amount")))),
                        0),
                    balance=Sum(Case(
                        When(is_credit=False, then=-F("amount")),
                        default=F("amount"))))]
    else:
        months = [utils.MonthlySummary(**x) for x in Record.objects
                  .filter(
                    Q(transaction__in=Transaction.objects.filter(
                        record__account__code__startswith=account.code)),
                    ~Q(account__code__startswith=account.code))
                  .annotate(month=TruncMonth("transaction__date"))
                  .values("month")
                  .order_by("month")
                  .annotate(
                    debit=Coalesce(
                        Sum(Case(When(is_credit=False, then=F("amount")))),
                        0),
                    credit=Coalesce(
                        Sum(Case(When(is_credit=True, then=F("amount")))),
                        0),
                    balance=Sum(Case(
                        When(is_credit=False, then=-F("amount")),
                        default=F("amount"))))]
    cumulative_balance = 0
    for month in months:
        cumulative_balance = cumulative_balance + month.balance
        month.cumulative_balance = cumulative_balance
    months.append(utils.MonthlySummary(
        label=_("Total"),
        credit=sum([x.credit for x in months]),
        debit=sum([x.debit for x in months]),
        balance=sum([x.balance for x in months]),
        cumulative_balance=cumulative_balance,
    ))
    try:
        pagination = Pagination(request, months, True)
    except PaginationException as e:
        return redirect(e.url)
    shortcut_accounts = utils.get_cash_shortcut_accounts()
    return render(request, "accounting/report-cash-summary.html", {
        "month_list": pagination.items,
        "pagination": pagination,
        "shortcut_accounts": [x for x in accounts if
                              x.code in shortcut_accounts],
        "all_accounts": [x for x in accounts if
                         x.code not in shortcut_accounts],
    })


@method_decorator(require_GET, name="dispatch")
class LedgerDefaultView(RedirectView):
    """The default ledger."""
    query_string = True
    pattern_name = "accounting:ledger"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["account"] = utils.get_default_ledger_account()
        kwargs["period"] = Period.default_spec()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def ledger(request: HttpRequest, account: Account,
           period: Period) -> HttpResponse:
    """The ledger.

    Args:
        request: The request.
        account: The account.
        period: The period.

    Returns:
        The response.
    """
    # The accounting records
    records = list(
        Record.objects
        .filter(
            transaction__date__gte=period.start,
            transaction__date__lte=period.end,
            account__code__startswith=account.code)
        .order_by("transaction__date", "transaction__ord", "is_credit",
                  "ord"))
    if re.match("^[1-3]", account.code) is not None:
        balance = Record.objects \
            .filter(
                transaction__date__lt=period.start,
                account__code__startswith=account.code) \
            .aggregate(
                balance=Coalesce(Sum(Case(When(
                    is_credit=True, then=-1),
                    default=1) * F("amount")), 0))["balance"]
        record_brought_forward = Record(
            transaction=Transaction(date=period.start),
            account=account,
            summary=_("Brought Forward"),
            is_credit=balance < 0,
            amount=abs(balance),
        )
        record_brought_forward.balance = balance
    else:
        balance = 0
        record_brought_forward = None
    for record in records:
        balance = balance + \
                  (-1 if record.is_credit else 1) * record.amount
        record.balance = balance
    if record_brought_forward is not None:
        records.insert(0, record_brought_forward)
    try:
        pagination = Pagination(request, records, True)
    except PaginationException as e:
        return redirect(e.url)
    records = pagination.items
    utils.find_imbalanced(records)
    utils.find_order_holes(records)
    utils.find_payable_records(account, records)
    utils.find_existing_equipments(account, records)
    return render(request, "accounting/report-ledger.html", {
        "record_list": records,
        "pagination": pagination,
        "accounts": utils.get_ledger_accounts(),
    })


@method_decorator(require_GET, name="dispatch")
class LedgerSummaryDefaultView(RedirectView):
    """The default ledger summary."""
    query_string = True
    pattern_name = "accounting:ledger-summary"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["account"] = utils.get_default_ledger_account()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def ledger_summary(request: HttpRequest, account: Account) -> HttpResponse:
    """The ledger summary report.

    Args:
        request: The request.
        account: The account.

    Returns:
        The response.
    """
    # The month summaries
    months = [utils.MonthlySummary(**x) for x in Record.objects
              .filter(account__code__startswith=account.code)
              .annotate(month=TruncMonth("transaction__date"))
              .values("month")
              .order_by("month")
              .annotate(
                    debit=Coalesce(
                        Sum(Case(When(is_credit=False, then=F("amount")))),
                        0),
                    credit=Coalesce(
                        Sum(Case(When(is_credit=True, then=F("amount")))),
                        0),
                    balance=Sum(Case(
                        When(is_credit=False, then=F("amount")),
                        default=-F("amount"))))]
    cumulative_balance = 0
    for month in months:
        cumulative_balance = cumulative_balance + month.balance
        month.cumulative_balance = cumulative_balance
    months.append(utils.MonthlySummary(
        label=_("Total"),
        credit=sum([x.credit for x in months]),
        debit=sum([x.debit for x in months]),
        balance=sum([x.balance for x in months]),
        cumulative_balance=cumulative_balance,
    ))
    try:
        pagination = Pagination(request, months, True)
    except PaginationException as e:
        return redirect(e.url)
    return render(request, "accounting/report-ledger-summary.html", {
        "month_list": pagination.items,
        "pagination": pagination,
        "accounts": utils.get_ledger_accounts(),
    })


@method_decorator(require_GET, name="dispatch")
class JournalDefaultView(RedirectView):
    """The default journal."""
    query_string = True
    pattern_name = "accounting:journal"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["period"] = Period.default_spec()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def journal(request: HttpRequest, period: Period) -> HttpResponse:
    """The journal.

    Args:
        request: The request.
        period: The period.

    Returns:
        The response.
    """
    # The accounting records
    records = Record.objects \
        .filter(
            transaction__date__gte=period.start,
            transaction__date__lte=period.end) \
        .order_by("transaction__date", "transaction__ord", "is_credit", "ord")
    # The brought-forward records
    brought_forward_accounts = Account.objects \
        .filter(
            Q(code__startswith="1")
            | Q(code__startswith="2")
            | Q(code__startswith="3")) \
        .annotate(balance=Sum(
            Case(
                When(record__is_credit=True, then=-1),
                default=1
            ) * F("record__amount"),
            filter=Q(record__transaction__date__lt=period.start))) \
        .filter(~Q(balance=0))
    debit_records = [Record(
        transaction=Transaction(date=period.start),
        account=x,
        is_credit=False,
        amount=x.balance
    ) for x in brought_forward_accounts if x.balance > 0]
    credit_records = [Record(
        transaction=Transaction(date=period.start),
        account=x,
        is_credit=True,
        amount=-x.balance
    ) for x in brought_forward_accounts if x.balance < 0]
    sum_debits = sum([x.amount for x in debit_records])
    sum_credits = sum([x.amount for x in credit_records])
    if sum_debits < sum_credits:
        debit_records.append(Record(
            transaction=Transaction(date=period.start),
            account=Account.objects.get(code=Account.ACCUMULATED_BALANCE),
            is_credit=False,
            amount=sum_credits - sum_debits
        ))
    elif sum_debits > sum_credits:
        credit_records.append(Record(
            transaction=Transaction(date=period.start),
            account=Account.objects.get(code=Account.ACCUMULATED_BALANCE),
            is_credit=True,
            amount=sum_debits - sum_credits
        ))
    records = list(debit_records) + list(credit_records) + list(records)
    try:
        pagination = Pagination(request, records, True)
    except PaginationException as e:
        return redirect(e.url)
    return render(request, "accounting/report-journal.html", {
        "record_list": pagination.items,
        "pagination": pagination,
    })


@method_decorator(require_GET, name="dispatch")
class TrialBalanceDefaultView(RedirectView):
    """The default trial balance."""
    query_string = True
    pattern_name = "accounting:trial-balance"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["period"] = Period.default_spec()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def trial_balance(request: HttpRequest, period: Period) -> HttpResponse:
    """The trial balance.

    Args:
        request: The request.
        period: The period.

    Returns:
        The response.
    """
    # The accounts
    nominal = list(
        Account.objects
        .filter(
            Q(record__transaction__date__gte=period.start),
            Q(record__transaction__date__lte=period.end),
            ~(Q(code__startswith="1")
              | Q(code__startswith="2")
              | Q(code__startswith="3")))
        .annotate(
            amount=Sum(Case(
                When(record__is_credit=True, then=-1),
                default=1) * F("record__amount")))
        .filter(Q(amount__isnull=False), ~Q(amount=0))
        .annotate(
            debit_amount=Case(
                When(amount__gt=0, then=F("amount")),
                default=None),
            credit_amount=Case(
                When(amount__lt=0, then=-F("amount")),
                default=None))
        .order_by("code"))
    real = list(
        Account.objects
        .filter(
            Q(record__transaction__date__lte=period.end),
            (Q(code__startswith="1")
             | Q(code__startswith="2")
             | Q(code__startswith="3")),
            ~Q(code=Account.ACCUMULATED_BALANCE))
        .annotate(
            amount=Sum(Case(
                When(record__is_credit=True, then=-1),
                default=1) * F("record__amount")))
        .filter(Q(amount__isnull=False), ~Q(amount=0))
        .annotate(
            debit_amount=Case(
                When(amount__gt=0, then=F("amount")),
                default=None),
            credit_amount=Case(
                When(amount__lt=0, then=-F("amount")),
                default=None))
        .order_by("code"))
    balance = Record.objects \
        .filter(
            (Q(transaction__date__lt=period.start)
             & ~(Q(account__code__startswith="1")
                 | Q(account__code__startswith="2")
                 | Q(account__code__startswith="3")))
            | (Q(transaction__date__lte=period.end)
               & Q(account__code=Account.ACCUMULATED_BALANCE))) \
        .aggregate(
            balance=Sum(Case(
                When(is_credit=True, then=-1),
                default=1) * F("amount")))["balance"]
    if balance is not None and balance != 0:
        brought_forward = Account.objects.get(
            code=Account.ACCUMULATED_BALANCE)
        if balance > 0:
            brought_forward.debit_amount = balance
            brought_forward.credit_amount = None
        else:
            brought_forward.debit_amount = None
            brought_forward.credit_amount = -balance
        real.append(brought_forward)
    accounts = nominal + real
    accounts.sort(key=lambda x: x.code)
    total_account = Account()
    total_account.title = _("Total")
    total_account.debit_amount = sum([x.debit_amount for x in accounts
                                      if x.debit_amount is not None])
    total_account.credit_amount = sum([x.credit_amount for x in accounts
                                       if x.credit_amount is not None])
    return render(request, "accounting/report-trial-balance.html", {
        "account_list": accounts,
        "total_item": total_account,
    })


@method_decorator(require_GET, name="dispatch")
class IncomeStatementDefaultView(RedirectView):
    """The default income statement."""
    query_string = True
    pattern_name = "accounting:income-statement"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["period"] = Period.default_spec()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def income_statement(request: HttpRequest, period: Period) -> HttpResponse:
    """The income statement.

    Args:
        request: The request.
        period: The period.

    Returns:
        The response.
    """
    # The accounts
    accounts = list(
        Account.objects
        .filter(
            Q(record__transaction__date__gte=period.start),
            Q(record__transaction__date__lte=period.end),
            ~(Q(code__startswith="1")
              | Q(code__startswith="2")
              | Q(code__startswith="3")))
        .annotate(
            amount=Sum(Case(
                When(record__is_credit=True, then=1),
                default=-1) * F("record__amount")))
        .filter(Q(amount__isnull=False), ~Q(amount=0))
        .order_by("code"))
    groups = list(Account.objects.filter(
        code__in=[x.code[:2] for x in accounts]))
    sections = list(Account.objects.filter(
        Q(code="4") | Q(code="5") | Q(code="6")
        | Q(code="7") | Q(code="8") | Q(code="9")).order_by("code"))
    cumulative_accounts = {
        "5": Account(title=_("Gross Income")),
        "6": Account(title=_("Operating Income")),
        "7": Account(title=_("Before Tax Income")),
        "8": Account(title=_("After Tax Income")),
        "9": Account.objects.get(code=Account.NET_CHANGE),
    }
    cumulative_total = 0
    for section in sections:
        section.groups = [x for x in groups
                          if x.code[:1] == section.code]
        for group in section.groups:
            group.details = [x for x in accounts
                             if x.code[:2] == group.code]
            group.amount = sum([x.amount
                                for x in group.details])
        section.amount = sum([x.amount for x in section.groups])
        cumulative_total = cumulative_total + section.amount
        if section.code in cumulative_accounts:
            section.cumulative_total \
                = cumulative_accounts[section.code]
            section.cumulative_total.amount = cumulative_total
        else:
            section.cumulative_total = None
        section.has_next = True
    sections[-1].has_next = False
    return render(request, "accounting/report-income-statement.html", {
        "section_list": sections,
    })


@method_decorator(require_GET, name="dispatch")
class BalanceSheetDefaultView(RedirectView):
    """The default balance sheet."""
    query_string = True
    pattern_name = "accounting:balance-sheet"

    def get_redirect_url(self, *args, **kwargs):
        kwargs["period"] = Period.default_spec()
        return super().get_redirect_url(*args, **kwargs)


@require_GET
def balance_sheet(request: HttpRequest, period: Period) -> HttpResponse:
    """The balance sheet.

    Args:
        request: The request.
        period: The period.

    Returns:
        The response.
    """
    # The accounts
    accounts = list(
        Account.objects
        .filter(
            Q(record__transaction__date__lte=period.end),
            (Q(code__startswith="1")
             | Q(code__startswith="2")
             | Q(code__startswith="3")),
            ~Q(code=Account.ACCUMULATED_BALANCE))
        .annotate(
            amount=Sum(Case(
                When(record__is_credit=True, then=-1),
                default=1) * F("record__amount")))
        .filter(Q(amount__isnull=False), ~Q(amount=0))
        .order_by("code"))
    for account in accounts:
        account.url = reverse("accounting:ledger", args=[account, period],
                              current_app=request.resolver_match.namespace)
    balance = Record.objects \
        .filter(
            Q(transaction__date__lt=period.start)
            & ~((Q(account__code__startswith="1")
                 | Q(account__code__startswith="2")
                 | Q(account__code__startswith="3"))
                & ~Q(account__code=Account.ACCUMULATED_BALANCE))) \
        .aggregate(
            balance=Sum(Case(
                When(is_credit=True, then=-1),
                default=1) * F("amount")))["balance"]
    if balance is not None and balance != 0:
        brought_forward = Account.objects.get(
            code=Account.ACCUMULATED_BALANCE)
        brought_forward.amount = balance
        brought_forward.url = reverse(
            "accounting:income-statement", args=[period.period_before()],
            current_app=request.resolver_match.namespace)
        accounts.append(brought_forward)
    balance = Record.objects \
        .filter(
            Q(transaction__date__gte=period.start)
            & Q(transaction__date__lte=period.end)
            & ~((Q(account__code__startswith="1")
                 | Q(account__code__startswith="2")
                 | Q(account__code__startswith="3"))
                & ~Q(account__code=Account.ACCUMULATED_BALANCE))) \
        .aggregate(
            balance=Sum(Case(
                When(is_credit=True, then=-1),
                default=1) * F("amount")))["balance"]
    if balance is not None and balance != 0:
        net_change = Account.objects.get(code=Account.NET_CHANGE)
        net_change.amount = balance
        net_change.url = reverse(
            "accounting:income-statement", args=[period],
            current_app=request.resolver_match.namespace)
        accounts.append(net_change)
    for account in [x for x in accounts if x.code[0] in "23"]:
        account.amount = -account.amount
    groups = list(Account.objects.filter(
        code__in=[x.code[:2] for x in accounts]))
    sections = list(Account.objects.filter(
        Q(code="1") | Q(code="2") | Q(code="3")).order_by("code"))
    for section in sections:
        section.groups = [x for x in groups
                          if x.code[:1] == section.code]
        for group in section.groups:
            group.details = [x for x in accounts
                             if x.code[:2] == group.code]
            group.amount = sum([x.amount
                                for x in group.details])
        section.amount = sum([x.amount for x in section.groups])
    by_code = {x.code: x for x in sections}
    return render(request, "accounting/report-balance-sheet.html", {
        "assets": by_code["1"],
        "liabilities": by_code["2"],
        "owners_equity": by_code["3"],
    })


@method_decorator(require_GET, name="dispatch")
class SearchListView(TemplateView):
    """The search."""
    template_name = "accounting/search.html"

    def get_context_data(self, **kwargs):
        records = self._get_records()
        try:
            pagination = Pagination(self.request, records, True)
        except PaginationException as e:
            return redirect(e.url)
        context = super().get_context_data(**kwargs)
        context["record_list"] = pagination.items
        context["pagination"] = pagination
        return context

    def _get_records(self) -> List[Record]:
        """Returns the search result."""
        query = self.request.GET.get("q")
        if query is None:
            return []
        terms = self._parse_search_terms(query)
        if len(terms) == 0:
            return []
        conditions = [self._get_conditions_for_term(x) for x in terms]
        if len(conditions) == 1:
            return Record.objects.filter(conditions[0])
        combined = conditions[0]
        for x in conditions[1:]:
            combined = combined & x
        return Record.objects.filter(combined)

    @staticmethod
    def _get_conditions_for_term(term: str) -> Q:
        """Returns the search conditions for a term.

        Args:
            term: The term.

        Returns:
            The search conditions for this term.
        """
        conditions =\
            Q(account__in=Account.objects.filter(
                Q(title_l10n__icontains=term)
                | Q(l10n_set__value__icontains=term)
                | Q(code=term)))\
            | Q(summary__icontains=term)\
            | Q(transaction__notes__icontains=term)
        if re.match("^[0-9]+(?:\\.[0-9]+)?$", term):
            conditions = conditions | Q(amount=Decimal(term))
        if re.match("^[1-9][0-8]{9}$", term):
            conditions = conditions\
                         | Q(pk=int(term))\
                         | Q(transaction__pk=int(term))\
                         | Q(account__pk=int(term))
        try:
            conditions = conditions | Q(transaction__date=parse_date(term))
        except ValueError:
            pass
        try:
            date = datetime.datetime.strptime(term, "%Y")
            conditions = conditions\
                         | Q(transaction__date__year=date.year)
        except ValueError:
            pass
        try:
            date = datetime.datetime.strptime(term, "%Y/%m")
            conditions = conditions\
                         | (Q(transaction__date__year=date.year)
                            & Q(transaction__date__month=date.month))
        except ValueError:
            pass
        try:
            date = datetime.datetime.strptime(term, "%m/%d")
            conditions = conditions\
                         | (Q(transaction__date__month=date.month)
                            & Q(transaction__date__day=date.day))
        except ValueError:
            pass
        return conditions

    @staticmethod
    def _parse_search_terms(query: str) -> List[str]:
        """Parses the search query and returns the search terms.  The search
        terms are separated by spaces but quoted with double quotes.

        Args:
            query: The search query

        Returns:
            The search terms.
        """
        query = query.strip()
        terms = []
        while True:
            m = re.match("^([^\"\\s]+)\\s*(.*)$", query)
            if m is not None:
                terms.append(m[1])
                query = m[2]
                continue
            m = re.match("^\"([^\"]*)\"\\s*(.*)$", query)
            if m is not None:
                if m[1] != "":
                    terms.append(m[1])
                query = m[2]
                continue
            m = re.match("^\"([^\"]*)", query)
            if m is not None:
                if m[1] != "":
                    terms.append(m[1])
            break
        return terms


@method_decorator(require_GET, name="dispatch")
class TransactionView(DetailView):
    """The view of the details of an accounting transaction."""
    context_object_name = "txn"

    def get_object(self, queryset=None):
        return self.kwargs["txn"]

    def get_template_names(self):
        model_name = self.object.__class__.__name__.lower()
        txn_type = self.kwargs["txn_type"]
        return [F"accounting/{model_name}_{txn_type}_detail.html"]


class TransactionFormView(FormView):
    """The form to create or update an accounting transaction."""
    model = Transaction
    form_class = TransactionForm
    not_modified_message = gettext_noop("This transaction was not modified.")
    success_message = gettext_noop("This transaction was saved successfully.")
    DEFAULT_REGULAR_ACCOUNTS = {
        "debit": [
            (gettext_noop("Pension"),
             gettext_noop("Pension for (last_month_name)"),
             "1314"),
            (gettext_noop("Health insurance"),
             gettext_noop("Health insurance for (last_month_name)"),
             "6262"),
            (gettext_noop("Electricity bill"),
             gettext_noop("Electricity bill for (last_bimonthly_from_name)"
                          "-(last_bimonthly_to_name)"),
             "6261"),
            (gettext_noop("Water bill"),
             gettext_noop("Water bill for (last_bimonthly_from_name)"
                          "-(last_bimonthly_to_name)"),
             "6261"),
            (gettext_noop("Gas bill"),
             gettext_noop("Gas bill for (last_bimonthly_from_name)"
                          "-(last_bimonthly_to_name)"),
             "6261"),
            (gettext_noop("Phone bill"),
             gettext_noop("Phone bill for (last_month_name)"),
             "6261"),
        ],
        "credit": [
            (gettext_noop("Payroll"),
             gettext_noop("Payroll for (last_month_name)"),
             "4611"),
        ],
    }

    def get_context_data(self, **kwargs):
        """Returns the context data for the template."""
        context = super().get_context_data(**kwargs)
        context["summary_categories"] = self._get_summary_categories()
        context["regular_accounts"] = self._get_regular_accounts()
        context["new_record_template"] = self._get_new_record_template_json()
        return context

    @staticmethod
    def _get_summary_categories() -> str:
        """Finds and returns the summary categories and their corresponding
        account hints as JSON.

        Returns:
            The summary categories and their account hints, by their record
            types and category types.
        """
        rows = Record.objects \
            .filter(Q(summary__contains="—"),
                    ~Q(account__code__startswith="114"),
                    ~Q(account__code__startswith="214"),
                    ~Q(account__code__startswith="128"),
                    ~Q(account__code__startswith="228")) \
            .annotate(rec_type=Case(When(is_credit=True, then=Value("credit")),
                                    default=Value("debit"),
                                    output_field=CharField()),
                      cat_type=Case(
                          When(summary__regex=".+—.+—.+→.+",
                               then=Value("bus")),
                          When(summary__regex=".+—.+[→↔].+",
                               then=Value("travel")),
                          default=Value("general"),
                          output_field=CharField()),
                      category=Left("summary",
                                    StrIndex("summary", Value("—")) - 1,
                                    output_field=CharField())) \
            .values("rec_type", "cat_type", "category", "account__code") \
            .annotate(count=Count("category")) \
            .order_by("rec_type", "cat_type", "category", "-count",
                      "account__code")
        # Sorts the rows by the record type and the category type
        categories = {}
        for row in rows:
            key = "%s-%s" % (row["rec_type"], row["cat_type"])
            if key not in categories:
                categories[key] = {}
            if row["category"] not in categories[key]:
                categories[key][row["category"]] = []
            categories[key][row["category"]].append(row)
        for key in categories:
            # Keeps only the first account with most records
            categories[key] = [categories[key][x][0] for x in categories[key]]
            # Sorts the categories by the frequency
            categories[key].sort(key=lambda x: (-x["count"], x["category"]))
            # Keeps only the category and the account
            categories[key] = [[x["category"], x["account__code"]]
                               for x in categories[key]]
        # Converts the dictionary to a list, as the category may not be
        # US-ASCII
        return json.dumps(categories)

    @staticmethod
    def _get_regular_accounts() -> str:
        """Returns the regular account data, as JSON.

        Returns:
            Two lists of the (title, format pattern, account code) tuple,
            sorted by debit or credit.
        """
        try:
            regular = settings.REGULAR_ACCOUNTS
            regular = {t: [{"title": x[0],
                            "format": x[1],
                            "account": x[2]} for x in regular[t]]
                       for t in regular}
        except AttributeError:
            regular = TransactionFormView.DEFAULT_REGULAR_ACCOUNTS
            regular = {t: [{"title": gettext(x[0]),
                            "format": gettext(x[1]),
                            "account": x[2]} for x in regular[t]]
                       for t in regular}
        return json.dumps(regular)

    def _get_new_record_template_json(self) -> str:
        context = {"record_type": "TTT", "no": "NNN"}
        template_name = "accounting/include/record_form-transfer.html"\
            if self.txn_type == "transfer"\
            else "accounting/include/record_form-non-transfer.html"
        return json.dumps(render_to_string(template_name, context))

    def get_template_name(self) -> str:
        """Returns the name of the template."""
        model_name = self.model.__name__.lower()
        return F"accounting/{model_name}_{self.txn_type}_form.html"

    def make_form_from_post(self, post: Dict[str, str]) -> TransactionForm:
        """Creates and returns the form from the POST data."""
        return TransactionForm.from_post(post, self.txn_type, self.object)

    def make_form_from_model(self, obj: Transaction) -> TransactionForm:
        """Creates and returns the form from a data model."""
        return TransactionForm.from_model(obj, self.txn_type)

    def fill_model_from_form(self, obj: Transaction,
                             form: TransactionForm) -> None:
        """Fills in the data model from the form."""
        obj.fill_from_post(form.data, self.request, self.txn_type)

    def get_object(self) -> Optional[Account]:
        """Returns the current object, or None on a create form."""
        return self.kwargs.get("txn")

    def get_success_url(self) -> str:
        """Returns the URL on success."""
        return reverse("accounting:transactions.detail",
                       args=[self.txn_type, self.object],
                       current_app=self.request.resolver_match.namespace)

    @property
    def txn_type(self) -> str:
        """Returns the transaction type of this form."""
        return self.kwargs["txn_type"]


@method_decorator(require_POST, name="dispatch")
class TransactionDeleteView(DeleteView):
    """The view to delete an accounting transaction."""
    success_message = gettext_noop(
            "This transaction was deleted successfully.")

    def get_object(self, queryset=None):
        return self.kwargs["txn"]

    def get_success_url(self):
        return self.request.GET.get("r")\
               or reverse("accounting:home",
                          current_app=self.request.resolver_match.namespace)


class TransactionSortFormView(FormView):
    """The form to sort the transactions in a same day."""
    template_name = "accounting/transaction_sort_form.html"
    form_class = TransactionSortForm
    not_modified_message = gettext_noop(
        "The transaction orders were not modified.")
    success_message = gettext_noop(
        "The transaction orders were saved successfully.")

    def get_form(self, **kwargs):
        """Returns the form for the template."""
        form = super().get_form()
        if form.txn_list is None:
            form.date = self.kwargs["date"]
            form.txn_list = Transaction.objects.filter(date=form.date)\
                .order_by("ord").all()
        if len(form.txn_list) < 2:
            raise Http404
        return form

    def make_form_from_post(self, post: Dict[str, str]) -> TransactionSortForm:
        """Creates and returns the form from the POST data."""
        return TransactionSortForm.from_post(self.kwargs["date"], post)

    def form_valid(self, form: TransactionSortForm) -> HttpResponseRedirect:
        """Handles the action when the POST form is valid."""
        modified = [x for x in form.txn_orders if x.txn.ord != x.order]
        if len(modified) == 0:
            message = self.get_not_modified_message(form.cleaned_data)
        else:
            with transaction.atomic():
                for x in modified:
                    Transaction.objects.filter(pk=x.txn.pk).update(ord=x.order)
            message = self.get_success_message(form.cleaned_data)
        messages.success(self.request, message)
        return redirect(self.get_success_url())

    def get_success_url(self) -> str:
        """Returns the URL on success."""
        return self.request.GET.get("r")\
            or reverse("accounting:home",
                       current_app=self.request.resolver_match.namespace)


@method_decorator(require_GET, name="dispatch")
class AccountListView(ListView):
    """The view to list the accounts."""
    queryset = Account.objects\
        .annotate(is_parent_and_in_use=ExpressionWrapper(
            Exists(Account.objects.filter(parent=OuterRef("pk")))
            & Exists(Record.objects.filter(account=OuterRef("pk"))),
            output_field=BooleanField()))\
        .order_by("code")


@method_decorator(require_GET, name="dispatch")
class AccountView(DetailView):
    """The view of an account."""
    def get_object(self, queryset=None):
        return self.kwargs["account"]


class AccountFormView(FormView):
    """The form to create or update an account."""
    model = Account
    form_class = AccountForm
    not_modified_message = gettext_noop("This account was not modified.")
    success_message = gettext_noop("This account was saved successfully.")

    def make_form_from_post(self, post: Dict[str, str]) -> AccountForm:
        """Creates and returns the form from the POST data."""
        form = AccountForm(post)
        form.account = self.object
        return form

    def make_form_from_model(self, obj: Account) -> AccountForm:
        """Creates and returns the form from a data model."""
        form = AccountForm({
            "code": obj.code,
            "title": obj.title,
        })
        form.account = obj
        return form

    def get_object(self) -> Optional[Account]:
        """Returns the current object, or None on a create form."""
        return self.kwargs.get("account")

    def get_success_url(self) -> str:
        """Returns the URL on success."""
        return reverse("accounting:accounts.detail", args=[self.object],
                       current_app=self.request.resolver_match.namespace)


@require_POST
def account_delete(request: HttpRequest,
                   account: Account) -> HttpResponseRedirect:
    """The view to delete an account.

    Args:
        request (HttpRequest): The request.
        account (Account): The account.

    Returns:
        HttpResponseRedirect: The response.
    """
    if account.is_in_use:
        message = gettext_noop("This account is in use.")
        messages.error(request, message)
        return redirect("accounting:accounts.detail", account)
    account.delete()
    message = gettext_noop("This account was deleted successfully.")
    messages.success(request, message)
    return redirect("accounting:accounts")


@require_GET
def api_account_list(request: HttpRequest) -> JsonResponse:
    """The API view to return all the accounts.

    Args:
        request: The request.

    Returns:
        The response.
    """
    return JsonResponse({x.code: x.title for x in Account.objects.all()})


@require_GET
def api_account_options(request: HttpRequest) -> JsonResponse:
    """The API view to return the account options.

    Args:
        request: The request.

    Returns:
        The response.
    """
    accounts = Account.objects\
        .annotate(children_count=Count("child_set"))\
        .filter(children_count=0)\
        .annotate(record_count=Count("record"))\
        .annotate(is_in_use=Case(
            When(record_count__gt=0, then=True),
            default=False,
            output_field=BooleanField()))\
        .order_by("code")
    for x in accounts:
        x.is_for_debit = re.match("^([1235689]|7[5678])", x.code) is not None
        x.is_for_credit = re.match("^([123489]|7[1234])", x.code) is not None
    return JsonResponse({
        "header_in_use": _("---Accounts In Use---"),
        "debit_in_use": [x.option_data for x in accounts
                         if x.is_for_debit and x.is_in_use],
        "credit_in_use": [x.option_data for x in accounts
                          if x.is_for_credit and x.is_in_use],
        "header_not_in_use": _("---Accounts Not In Use---"),
        "debit_not_in_use": [x.option_data for x in accounts
                             if x.is_for_debit and not x.is_in_use],
        "credit_not_in_use": [x.option_data for x in accounts
                              if x.is_for_credit and not x.is_in_use],
    })



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/6/30

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The period chooser utilities of the Mia core application.

"""
import datetime
import re
from typing import Optional, List

from django.core.serializers.json import DjangoJSONEncoder
from django.template import defaultfilters
from django.utils import dateformat, timezone
from django.utils.translation import gettext

from mia_core.utils import Language


class Period:
    """The template helper for the period chooser.

    Args:
        spec: The current period specification
        data_start: The available first day of the data.
        data_end: The available last day of the data.

    Raises:
        ValueError: When the period specification is invalid.
    """
    def __init__(self, spec: str = None, data_start: datetime.date = None,
                 data_end: datetime.date = None):
        # Raises ValueError
        self._period = self.Parser(spec)
        self._data_start = data_start
        self._data_end = data_end

    @property
    def spec(self) -> str:
        """Returns the period specification.

        Returns:
            The period specification.
        """
        return self._period.spec

    @property
    def start(self) -> datetime.date:
        """Returns the start day of the currently-specified period.

        Returns:
            The start day of the currently-specified period.
        """
        return self._period.start

    @property
    def end(self) -> datetime.date:
        """Returns the end day of the currently-specified period.

        Returns:
            The end day of the currently-specified period.
        """
        return self._period.end

    @property
    def description(self) -> str:
        """Returns the text description of the currently-specified period.

        Returns:
            The text description of the currently-specified period
        """
        return self._period.description

    @property
    def prep_desc(self) -> str:
        """Returns the text description with preposition of the
        currently-specified period.

        Returns:
            The text description with preposition of the currently-specified
            period.
        """
        return self._period.prep_desc

    @staticmethod
    def _get_last_month_start() -> datetime.date:
        """Returns the first day of the last month.

        Returns:
            The first day of the last month.
        """
        today = timezone.localdate()
        month = today.month - 1
        year = today.year
        if month < 1:
            month = 12
            year = year - 1
        return datetime.date(year, month, 1)

    @staticmethod
    def _get_next_month_start() -> datetime.date:
        """Returns the first day of the next month.

        Returns:
            The first day of the next month.
        """
        today = timezone.localdate()
        month = today.month + 1
        year = today.year
        if month > 12:
            month = 1
            year = year + 1
        return datetime.date(year, month, 1)

    def this_month(self) -> Optional[str]:
        """Returns the specification of this month.

        Returns:
            The specification of this month, or None if there is no data in or
            before this month.
        """
        if self._data_start is None:
            return None
        today = timezone.localdate()
        first_month_start = datetime.date(
            self._data_start.year, self._data_start.month, 1)
        if today < first_month_start:
            return None
        return dateformat.format(today, "Y-m")

    def last_month(self) -> Optional[str]:
        """Returns the specification of last month.

        Returns:
            The specification of last month, or None if there is no data in or
            before last month.
        """
        if self._data_start is None:
            return None
        last_month_start = self._get_last_month_start()
        first_month_start = datetime.date(
            self._data_start.year, self._data_start.month, 1)
        if last_month_start < first_month_start:
            return None
        return dateformat.format(last_month_start, "Y-m")

    def since_last_month(self) -> Optional[str]:
        """Returns the specification since last month.

        Returns:
            The specification since last month, or None if there is no data in
            or before last month.
        """
        last_month = self.last_month()
        if last_month is None:
            return None
        return last_month + "-"

    def has_months_to_choose(self) -> bool:
        """Returns whether there are months to choose besides this month and
        last month.

        Returns:
            True if there are months to choose besides this month and last
            month, or False otherwise.
        """
        if self._data_start is None:
            return False
        if self._data_start < self._get_last_month_start():
            return True
        if self._data_end >= self._get_next_month_start():
            return True
        return False

    def chosen_month(self) -> Optional[str]:
        """Returns the specification of the chosen month, or None if the
        current period is not a month or is out of available data range.

        Returns:
            The specification of the chosen month, or None if the current
            period is not a month or is out of available data range.
        """
        if self._data_start is None:
            return None
        m = re.match("^[0-9]{4}-[0-2]{2}", self._period.spec)
        if m is None:
            return None
        if self._period.end < self._data_start:
            return None
        if self._period.start > self._data_end:
            return None
        return self._period.spec

    def this_year(self) -> Optional[str]:
        """Returns the specification of this year.

        Returns:
            The specification of this year, or None if there is no data in or
            before this year.
        """
        if self._data_start is None:
            return None
        this_year = timezone.localdate().year
        if this_year < self._data_start.year:
            return None
        return str(this_year)

    def last_year(self) -> Optional[str]:
        """Returns the specification of last year.

        Returns:
            The specification of last year, or None if there is no data in or
            before last year.
        """
        if self._data_start is None:
            return None
        last_year = timezone.localdate().year - 1
        if last_year < self._data_start.year:
            return None
        return str(last_year)

    def has_years_to_choose(self) -> bool:
        """Returns whether there are years to choose besides this year and
        last year.

        Returns:
            True if there are years to choose besides this year and last year,
            or False otherwise.
        """
        if self._data_start is None:
            return False
        this_year = timezone.localdate().year
        if self._data_start.year < this_year - 1:
            return True
        if self._data_end.year > this_year:
            return True
        return False

    def years_to_choose(self) -> Optional[List[str]]:
        """Returns the years to choose besides this year and last year.

        Returns:
            The years to choose besides this year and last year, or None if
            there is no data.
        """
        if self._data_start is None:
            return None
        this_year = timezone.localdate().year
        before = [str(x) for x in range(
            self._data_start.year, this_year - 1)]
        after = [str(x) for x in range(
            self._data_end.year, this_year, -1)]
        return after + before[::-1]

    def today(self) -> Optional[None]:
        """Returns the specification of today.

        Returns:
            The specification of today, or None if there is no data in or
            before today.
        """
        if self._data_start is None:
            return None
        today = timezone.localdate()
        if today < self._data_start or today > self._data_end:
            return None
        return dateformat.format(today, "Y-m-d")

    def yesterday(self) -> Optional[str]:
        """Returns the specification of yesterday.

        Returns:
            The specification of yesterday, or None if there is no data in or
            before yesterday.
        """
        if self._data_start is None:
            return None
        yesterday = timezone.localdate() - datetime.timedelta(days=1)
        if yesterday < self._data_start or yesterday > self._data_end:
            return None
        return dateformat.format(yesterday, "Y-m-d")

    def chosen_day(self) -> str:
        """Returns the specification of the chosen day.

        Returns:
            The specification of the chosen day, or the start day of the period
            if the current period is not a day.
        """
        return dateformat.format(self._period.start, "Y-m-d")

    def has_days_to_choose(self) -> bool:
        """Returns whether there are more than one day to choose from.

        Returns:
            True if there are more than one day to choose from, or False
            otherwise.
        """
        if self._data_start is None:
            return False
        if self._data_start == self._data_end:
            return False
        return True

    def first_day(self) -> Optional[str]:
        """Returns the specification of the available first day.

        Returns:
            The specification of the available first day, or None if there is
            no data.
        """
        if self._data_start is None:
            return None
        return dateformat.format(self._data_start, "Y-m-d")

    def last_day(self) -> Optional[str]:
        """Returns the specification of the available last day.

        Returns:
            The specification of the available last day, or None if there is no
            data.
        """
        if self._data_end is None:
            return None
        return dateformat.format(self._data_end, "Y-m-d")

    def chosen_start(self) -> Optional[str]:
        """Returns the specification of of the first day of the
        specified period.

        Returns:
            The specification of of the first day of the specified period, or
            None if there is no data.
        """
        if self._data_start is None:
            return None
        day = self._period.start \
            if self._period.start >= self._data_start \
            else self._data_start
        return dateformat.format(day, "Y-m-d")

    def chosen_end(self) -> Optional[str]:
        """Returns the specification of of the last day of the
        specified period.

        Returns:
            The specification of of the last day of the specified period, or
            None if there is data.
        """
        if self._data_end is None:
            return None
        day = self._period.end \
            if self._period.end <= self._data_end \
            else self._data_end
        return dateformat.format(day, "Y-m-d")

    def period_before(self) -> Optional[str]:
        """Returns the specification of the period before the current period.

        Returns:
            The specification of the period before the current period, or None
            if there is no data before the current period.
        """
        if self._data_start is None:
            return None
        if self.start <= self._data_start:
            return None
        previous_day = self.start - datetime.timedelta(days=1)
        if re.match("^[0-9]{4}$", self.spec):
            return "-" + str(previous_day.year)
        if re.match("^[0-9]{4}-[0-9]{2}$", self.spec):
            return dateformat.format(previous_day, "-Y-m")
        return dateformat.format(previous_day, "-Y-m-d")

    def month_picker_params(self) -> Optional[str]:
        """Returns the parameters for the month-picker, as a JSON text string.

        Returns:
            The parameters for the month-picker, as a JSON text string, or None
            if there is no data.
        """
        if self._data_start is None:
            return None
        start = datetime.date(self._data_start.year, self._data_start.month, 1)
        return DjangoJSONEncoder().encode({
            "locale": Language.current().locale,
            "minDate": start,
            "maxDate": self._data_end,
            "defaultDate": self.chosen_month(),
        })

    @staticmethod
    def default_spec() -> str:
        """Returns the specification for the default period.

        Returns:
            str: The specification for the default period
        """
        return dateformat.format(timezone.localdate(), "Y-m")

    class Parser:
        """The period parser.

        Args:
            spec (str|None): The period specification.

        Raises:
            ValueError: When the period specification is invalid.

        Attributes:
            spec (str): The currently-using period specification.
            start (datetime.date): The start of the period.
            end (datetime.date): The end of the period.
            description (str): The text description of the period.
            prep_desc (str): The text description with preposition.
        """
        VERY_START: datetime.date = datetime.date(1990, 1, 1)

        def __init__(self, spec: str):
            self.spec = None
            self.start = None
            self.end = None
            self.description = None
            self.prep_desc = None

            if spec is None:
                self._set_this_month()
                return
            self.spec = spec
            # A specific month
            m = re.match("^([0-9]{4})-([0-9]{2})$", spec)
            if m is not None:
                year = int(m.group(1))
                month = int(m.group(2))
                # Raises ValueError
                self.start = datetime.date(year, month, 1)
                self.end = self._month_last_day(self.start)
                self.description = self._month_text(year, month)
                self.prep_desc = gettext("In %s") % self.description
                return
            # From a specific month
            m = re.match("^([0-9]{4})-([0-9]{2})-$", spec)
            if m is not None:
                year = int(m.group(1))
                month = int(m.group(2))
                # Raises ValueError
                self.start = datetime.date(year, month, 1)
                self.end = self._month_last_day(timezone.localdate())
                self.description = gettext("Since %s")\
                    % self._month_text(year, month)
                self.prep_desc = self.description
                return
            # Until a specific month
            m = re.match("^-([0-9]{4})-([0-9]{2})$", spec)
            if m is not None:
                year = int(m.group(1))
                month = int(m.group(2))
                # Raises ValueError
                until_month = datetime.date(year, month, 1)
                self.start = Period.Parser.VERY_START
                self.end = self._month_last_day(until_month)
                self.description = gettext("Until %s")\
                    % self._month_text(year, month)
                self.prep_desc = self.description
                return
            # A specific year
            m = re.match("^([0-9]{4})$", spec)
            if m is not None:
                year = int(m.group(1))
                # Raises ValueError
                self.start = datetime.date(year, 1, 1)
                self.end = datetime.date(year, 12, 31)
                self.description = self._year_text(year)
                self.prep_desc = gettext("In %s") % self.description
                return
            # Until a specific year
            m = re.match("^-([0-9]{4})$", spec)
            if m is not None:
                year = int(m.group(1))
                # Raises ValueError
                self.end = datetime.date(year, 12, 31)
                self.start = Period.Parser.VERY_START
                self.description = gettext("Until %s")\
                    % self._year_text(year)
                self.prep_desc = self.description
                return
            # All time
            if spec == "-":
                self.start = Period.Parser.VERY_START
                self.end = self._month_last_day(timezone.localdate())
                self.description = gettext("All Time")
                self.prep_desc = gettext("In %s") % self.description
                return
            # A specific date
            m = re.match("^([0-9]{4})-([0-9]{2})-([0-9]{2})$",
                         spec)
            if m is not None:
                # Raises ValueError
                self.start = datetime.date(
                    int(m.group(1)),
                    int(m.group(2)),
                    int(m.group(3)))
                self.end = self.start
                self.description = self._date_text(self.start)
                self.prep_desc = gettext("In %s") % self.description
                return
            # A specific date period
            m = re.match(("^([0-9]{4})-([0-9]{2})-([0-9]{2})"
                          "-([0-9]{4})-([0-9]{2})-([0-9]{2})$"),
                         spec)
            if m is not None:
                # Raises ValueError
                self.start = datetime.date(
                    int(m.group(1)),
                    int(m.group(2)),
                    int(m.group(3)))
                self.end = datetime.date(
                    int(m.group(4)),
                    int(m.group(5)),
                    int(m.group(6)))
                today = timezone.localdate()
                # Spans several years
                if self.start.year != self.end.year:
                    self.description = "%s-%s" % (
                        defaultfilters.date(self.start, "Y/n/j"),
                        defaultfilters.date(self.end, "Y/n/j"))
                # Spans several months
                elif self.start.month != self.end.month:
                    if self.start.year != today.year:
                        self.description = "%s-%s" % (
                            defaultfilters.date(self.start, "Y/n/j"),
                            defaultfilters.date(self.end, "n/j"))
                    else:
                        self.description = "%s-%s" % (
                            defaultfilters.date(self.start, "n/j"),
                            defaultfilters.date(self.end, "n/j"))
                # Spans several days
                elif self.start.day != self.end.day:
                    if self.start.year != today.year:
                        self.description = "%s-%s" % (
                            defaultfilters.date(self.start, "Y/n/j"),
                            defaultfilters.date(self.end, "j"))
                    else:
                        self.description = "%s-%s" % (
                            defaultfilters.date(self.start, "n/j"),
                            defaultfilters.date(self.end, "j"))
                # At the same day
                else:
                    self.spec = dateformat.format(self.start, "Y-m-d")
                    self.description = self._date_text(self.start)
                self.prep_desc = gettext("In %s") % self.description
                return
            # Until a specific day
            m = re.match("^-([0-9]{4})-([0-9]{2})-([0-9]{2})$", spec)
            if m is not None:
                # Raises ValueError
                self.end = datetime.date(
                    int(m.group(1)),
                    int(m.group(2)),
                    int(m.group(3)))
                self.start = Period.Parser.VERY_START
                self.description = gettext("Until %s")\
                    % self._date_text(self.end)
                self.prep_desc = self.description
                return
            # Wrong period format
            raise ValueError

        def _set_this_month(self) -> None:
            """Sets the period to this month."""
            today = timezone.localdate()
            self.spec = dateformat.format(today, "Y-m")
            self.start = datetime.date(today.year, today.month, 1)
            self.end = self._month_last_day(self.start)
            self.description = gettext("This Month")

        @staticmethod
        def _month_last_day(day: datetime.date) -> datetime.date:
            """Calculates and returns the last day of a month.

            Args:
                day: A day in the month.

            Returns:
                The last day in the month
            """
            next_month = day.month + 1
            next_year = day.year
            if next_month > 12:
                next_month = 1
                next_year = next_year + 1
            return datetime.date(
                next_year, next_month, 1) - datetime.timedelta(days=1)

        @staticmethod
        def _month_text(year: int, month: int) -> str:
            """Returns the text description of a month.

            Args:
                year: The year.
                month: The month.

            Returns:
                The description of the month.
            """
            today = timezone.localdate()
            if year == today.year and month == today.month:
                return gettext("This Month")
            prev_month = today.month - 1
            prev_year = today.year
            if prev_month < 1:
                prev_month = 12
                prev_year = prev_year - 1
            prev = datetime.date(prev_year, prev_month, 1)
            if year == prev.year and month == prev.month:
                return gettext("Last Month")
            return "%d/%d" % (year, month)

        @staticmethod
        def _year_text(year: int) -> str:
            """Returns the text description of a year.

            Args:
                year: The year.

            Returns:
                The description of the year.
            """
            this_year = timezone.localdate().year
            if year == this_year:
                return gettext("This Year")
            if year == this_year - 1:
                return gettext("Last Year")
            return str(year)

        @staticmethod
        def _date_text(day: datetime.date) -> str:
            """Returns the text description of a day.

            Args:
                day: The date.

            Returns:
                The description of the day.
            """
            today = timezone.localdate()
            if day == today:
                return gettext("Today")
            elif day == today - datetime.timedelta(days=1):
                return gettext("Yesterday")
            elif day.year != today.year:
                return defaultfilters.date(day, "Y/n/j")
            else:
                return defaultfilters.date(day, "n/j")



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/31

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The forms of the Mia core application.

"""
import datetime
import re
from decimal import Decimal
from typing import Optional, List, Dict

from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import RegexValidator
from django.db.models import Q, Max, Model
from django.db.models.functions import Length
from django.utils.translation import gettext as _

from .models import Account, Record, Transaction
from .validators import validate_record_account_code, validate_record_id


class RecordForm(forms.Form):
    """An accounting record form.

    Attributes:
        txn_form (TransactionForm): The parent transaction form.
        is_credit (bool): Whether this is a credit record.
    """
    id = forms.IntegerField(
        required=False,
        error_messages={
            "invalid": _("This accounting record is not valid."),
        },
        validators=[validate_record_id])
    account = forms.CharField(
        error_messages={
            "required": _("Please select the account."),
        },
        validators=[validate_record_account_code])
    summary = forms.CharField(
        required=False,
        max_length=128,
        error_messages={
            "max_length": _("This summary is too long (max. 128 characters)."),
        })
    amount = forms.DecimalField(
        max_digits=18,
        decimal_places=2,
        min_value=0.01,
        error_messages={
            "required": _("Please fill in the amount."),
            "invalid": _("Please fill in a number."),
            "min_value": _("The amount must be more than 0."),
        })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.txn_form = None
        self.is_credit = None

    def account_title(self) -> Optional[str]:
        """Returns the title of the specified account, if any.

        Returns:
            The title of the specified account, or None if the specified
            account is not available.
        """
        try:
            return Account.objects.get(code=self["account"].value()).title
        except KeyError:
            return None
        except Account.DoesNotExist:
            return None

    def clean(self):
        """Validates the form globally.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        errors = []
        if "id" in self.errors:
            errors = errors + self.errors["id"].as_data()
        validators = [self._validate_transaction,
                      self._validate_account_type,
                      self._validate_is_credit]
        for validator in validators:
            try:
                validator()
            except forms.ValidationError as e:
                errors.append(e)
        if errors:
            raise forms.ValidationError(errors)

    def _validate_transaction(self) -> None:
        """Validates whether the transaction matches the transaction form.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if "id" in self.errors:
            return
        if self.txn_form.transaction is None:
            if "id" in self.data:
                raise forms.ValidationError(
                    _("This record is not for this transaction."),
                    code="not_belong")
        else:
            if "id" in self.data:
                record = Record.objects.get(pk=self.data["id"])
                if record.transaction.pk != self.txn_form.transaction.pk:
                    raise forms.ValidationError(
                        _("This record is not for this transaction."),
                        code="not_belong")

    def _validate_account_type(self) -> None:
        """Validates whether the account is a correct debit or credit account.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if "account" in self.errors:
            return
        if self.is_credit:
            if not re.match("^([123489]|7[1234])", self.data["account"]):
                error = forms.ValidationError(
                    _("This account is not for credit records."),
                    code="not_credit")
                self.add_error("account", error)
                raise error
        else:
            if not re.match("^([1235689]|7[5678])", self.data["account"]):
                error = forms.ValidationError(
                    _("This account is not for debit records."),
                    code="not_debit")
                self.add_error("account", error)
                raise error

    def _validate_is_credit(self) -> None:
        """Validates whether debit and credit records are submitted correctly
        as corresponding debit and credit records.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if "id" in self.errors:
            return
        if "id" not in self.data:
            return
        record = Record.objects.get(pk=self.data["id"])
        if record.is_credit != self.is_credit:
            if self.is_credit:
                raise forms.ValidationError(
                    _("This accounting record is not a credit record."),
                    code="not_credit")
            else:
                raise forms.ValidationError(
                    _("This accounting record is not a debit record."),
                    code="not_debit")


class TransactionForm(forms.Form):
    """A transaction form.

    Attributes:
        txn_type (str): The transaction type.
        transaction (Transaction|None): The current transaction or None
        debit_records (list[RecordForm]): The debit records.
        credit_records (list[RecordForm]): The credit records.
    """
    date = forms.DateField(
        required=True,
        error_messages={
            "required": _("Please fill in the date."),
            "invalid": _("This date is not valid.")
        })
    notes = forms.CharField(
        required=False,
        max_length=128,
        error_messages={
            "max_length": _("These notes are too long (max. 128 characters).")
        })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Populates the belonging record forms
        self.debit_records = []
        self.credit_records = []
        if len(args) > 0 and isinstance(args[0], dict):
            by_rec_id = {}
            for key in args[0].keys():
                m = re.match(
                    ("^((debit|credit)-([1-9][0-9]*))-"
                     "(id|ord|account|summary|amount)$"),
                    key)
                if m is None:
                    continue
                rec_id = m.group(1)
                column = m.group(4)
                if rec_id not in by_rec_id:
                    by_rec_id[rec_id] = {
                        "is_credit": m.group(2) == "credit",
                        "no": int(m.group(3)),
                        "data": {},
                    }
                by_rec_id[rec_id]["data"][column] = args[0][key]
            debit_data_list = [x for x in by_rec_id.values()
                               if not x["is_credit"]]
            debit_data_list.sort(key=lambda x: x["no"])
            for x in debit_data_list:
                record_form = RecordForm(x["data"])
                record_form.txn_form = self
                record_form.is_credit = False
                self.debit_records.append(record_form)
            credit_data_list = [x for x in by_rec_id.values()
                                if x["is_credit"]]
            credit_data_list.sort(key=lambda x: x["no"])
            for x in credit_data_list:
                record_form = RecordForm(x["data"])
                record_form.txn_form = self
                record_form.is_credit = True
                self.credit_records.append(record_form)
        self.txn_type = None
        self.transaction = None

    @staticmethod
    def from_post(post: Dict[str, str], txn_type: str, txn: Model):
        """Constructs a transaction form from the POST data.

        Args:
            post: The post data.
            txn_type: The transaction type.
            txn: The transaction data model.

        Returns:
            The transaction form.
        """
        TransactionForm._sort_post_txn_records(post)
        form = TransactionForm(post)
        form.txn_type = txn_type
        form.transaction = txn
        return form

    @staticmethod
    def _sort_post_txn_records(post: Dict[str, str]) -> None:
        """Sorts the records in the form by their specified order, so that the
        form can be used to populate the data to return to the user.

        Args:
            post: The POSTed form.
        """
        # Collects the available record numbers
        record_no = {
            "debit": [],
            "credit": [],
        }
        for key in post.keys():
            m = re.match(
                "^(debit|credit)-([1-9][0-9]*)-(id|ord|account|summary|amount)",
                key)
            if m is None:
                continue
            record_type = m.group(1)
            no = int(m.group(2))
            if no not in record_no[record_type]:
                record_no[record_type].append(no)
        # Sorts these record numbers by their specified orders
        for record_type in record_no.keys():
            orders = {}
            for no in record_no[record_type]:
                try:
                    orders[no] = int(post[F"{record_type}-{no}-ord"])
                except KeyError:
                    orders[no] = 9999
                except ValueError:
                    orders[no] = 9999
            record_no[record_type].sort(key=lambda n: orders[n])
        # Constructs the sorted new form
        new_post = {}
        for record_type in record_no.keys():
            for i in range(len(record_no[record_type])):
                old_no = record_no[record_type][i]
                no = i + 1
                new_post[F"{record_type}-{no}-ord"] = str(no)
                for attr in ["id", "account", "summary", "amount"]:
                    if F"{record_type}-{old_no}-{attr}" in post:
                        new_post[F"{record_type}-{no}-{attr}"] \
                            = post[F"{record_type}-{old_no}-{attr}"]
        # Purges the old form and fills it with the new form
        for x in [x for x in post.keys() if re.match(
                "^(debit|credit)-([1-9][0-9]*)-(id|ord|account|summary|amount)",
                x)]:
            del post[x]
        for key in new_post.keys():
            post[key] = new_post[key]

    @staticmethod
    def from_model(txn: Transaction, txn_type: str):
        """Constructs a transaction form from the transaction data model.

        Args:
            txn: The transaction data model.
            txn_type: The transaction type.

        Returns:
            The transaction form.
        """
        form = TransactionForm(
            {x: str(getattr(txn, x)) for x in ["date", "notes"]
             if getattr(txn, x) is not None})
        form.transaction = txn if txn.pk is not None else None
        form.txn_type = txn_type
        records = []
        if txn_type != "income":
            records = records + txn.debit_records
        if txn_type != "expense":
            records = records + txn.credit_records
        for record in records:
            data = {x: getattr(record, x)
                    for x in ["summary", "amount"]
                    if getattr(record, x) is not None}
            if record.pk is not None:
                data["id"] = record.pk
            try:
                data["account"] = record.account.code
            except ObjectDoesNotExist:
                pass
            record_form = RecordForm(data)
            record_form.txn_form = form
            record_form.is_credit = record.is_credit
            if record.is_credit:
                form.credit_records.append(record_form)
            else:
                form.debit_records.append(record_form)
        return form

    def clean(self):
        """Validates the form globally.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        errors = []
        validators = [self._validate_has_debit_records,
                      self._validate_has_credit_records,
                      self._validate_balance]
        for validator in validators:
            try:
                validator()
            except forms.ValidationError as e:
                errors.append(e)
        if errors:
            raise forms.ValidationError(errors)

    def _validate_has_debit_records(self) -> None:
        """Validates whether there is any debit record.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if self.txn_type == "income":
            return
        if len(self.debit_records) > 0:
            return
        if self.txn_type == "transfer":
            raise forms.ValidationError(
                _("Please fill in debit accounting records."),
                code="has_debit_records")
        raise forms.ValidationError(
            _("Please fill in accounting records."),
            code="has_debit_records")

    def _validate_has_credit_records(self) -> None:
        """Validates whether there is any credit record.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if self.txn_type == "expense":
            return
        if len(self.credit_records) > 0:
            return
        if self.txn_type == "transfer":
            raise forms.ValidationError(
                _("Please fill in credit accounting records."),
                code="has_debit_records")
        raise forms.ValidationError(
            _("Please fill in accounting records."),
            code="has_debit_records")

    def _validate_balance(self) -> None:
        """Validates whether the total amount of debit and credit records are
        consistent.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if self.txn_type != "transfer":
            return
        if self.debit_total() == self.credit_total():
            return
        raise forms.ValidationError(
            _("The total of the debit and credit amounts are inconsistent."),
            code="balance")

    def is_valid(self) -> bool:
        if not super().is_valid():
            return False
        for x in self.debit_records + self.credit_records:
            if not x.is_valid():
                return False
        return True

    def balance_error(self) -> Optional[str]:
        """Returns the error message when the transaction is imbalanced.

        Returns:
            The error message when the transaction is imbalanced, or None
            otherwise.
        """
        errors = [x for x in self.non_field_errors().data
                  if x.code == "balance"]
        if errors:
            return errors[0].message
        return None

    def debit_total(self) -> Decimal:
        """Returns the total amount of the debit records.

        Returns:
            The total amount of the credit records.
        """
        return sum([Decimal(x.data["amount"]) for x in self.debit_records
                    if "amount" in x.data and "amount" not in x.errors])

    def credit_total(self) -> Decimal:
        """Returns the total amount of the credit records.

        Returns:
            The total amount of the credit records.
        """
        return sum([Decimal(x.data["amount"]) for x in self.credit_records
                    if "amount" in x.data and "amount" not in x.errors])


class TransactionSortForm(forms.Form):
    """A form to sort the transactions in a same day."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.date = None
        self.txn_list: Optional[List[Transaction]] = None
        self.txn_orders: List[TransactionSortForm.Order] = []

    @staticmethod
    def from_post(date: datetime.date, post: Dict[str, str]):
        form = TransactionSortForm({})
        form.date = date
        post_orders: List[TransactionSortForm.Order] = []
        for txn in Transaction.objects.filter(date=date).all():
            key = F"transaction-{txn.pk}-ord"
            if key not in post:
                post_orders.append(form.Order(txn, 9999))
            elif not re.match("^[0-9]+$", post[key]):
                post_orders.append(form.Order(txn, 9999))
            else:
                post_orders.append(form.Order(txn, int(post[key])))
        post_orders.sort(key=lambda x: (x.order, x.txn.ord))
        form.txn_orders = []
        for i in range(len(post_orders)):
            form.txn_orders.append(form.Order(post_orders[i].txn, i + 1))
        form.txn_list = [x.txn for x in form.txn_orders]
        return form

    class Order:
        """A transaction order"""
        def __init__(self, txn: Transaction, order: int):
            self.txn = txn
            self.order = order


class AccountForm(forms.Form):
    """An account form."""
    code = forms.CharField(
        error_messages={
            "required": _("Please fill in the code."),
            "invalid": _("Please fill in a number."),
            "max_length": _("This code is too long  (max. 5)."),
            "min_value": _("This code is too long  (max. 5)."),
        }, validators=[
            RegexValidator(
                regex="^[1-9]+$",
                message=_("You can only use numbers 1-9 in the code.")),
        ])
    title = forms.CharField(
        max_length=128,
        error_messages={
            "required": _("Please fill in the title."),
            "max_length": _("This title is too long  (max. 128)."),
        })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.account = None

    @property
    def parent(self) -> Optional[Account]:
        """The parent account, or None if this is the topmost account."""
        code = self["code"].value()
        if code is None or len(code) < 2:
            return None
        return Account.objects.get(code=code[:-1])

    def clean(self):
        """Validates the form globally.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        errors = []
        validators = [self._validate_code_not_under_myself,
                      self._validate_code_unique,
                      self._validate_code_parent_exists,
                      self._validate_code_descendant_code_size]
        for validator in validators:
            try:
                validator()
            except forms.ValidationError as e:
                errors.append(e)
        if errors:
            raise forms.ValidationError(errors)

    def _validate_code_not_under_myself(self) -> None:
        """Validates whether the code is under itself.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if self.account is None:
            return
        if "code" not in self.data:
            return
        if self.data["code"] == self.account.code:
            return
        if not self.data["code"].startswith(self.account.code):
            return
        error = forms.ValidationError(
            _("You cannot set the code under itself."),
            code="not_under_myself")
        self.add_error("code", error)
        raise error

    def _validate_code_unique(self) -> None:
        """Validates whether the code is unique.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if "code" not in self.data:
            return
        try:
            if self.account is None:
                Account.objects.get(code=self.data["code"])
            else:
                Account.objects.get(Q(code=self.data["code"]),
                                    ~Q(pk=self.account.pk))
        except Account.DoesNotExist:
            return
        error = forms.ValidationError(_("This code is already in use."),
                                      code="code_unique")
        self.add_error("code", error)
        raise error

    def _validate_code_parent_exists(self) -> None:
        """Validates whether the parent account exists.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if "code" not in self.data:
            return
        if len(self.data["code"]) < 2:
            return
        try:
            Account.objects.get(code=self.data["code"][:-1])
        except Account.DoesNotExist:
            error = forms.ValidationError(
                _("The parent account of this code does not exist."),
                code="code_unique")
            self.add_error("code", error)
            raise error
        return

    def _validate_code_descendant_code_size(self) -> None:
        """Validates whether the codes of the descendants will be too long.

        Raises:
            forms.ValidationError: When the validation fails.
        """
        if "code" not in self.data:
            return
        if self.account is None:
            return
        cur_max_len = Account.objects\
            .filter(Q(code__startswith=self.account.code),
                    ~Q(pk=self.account.pk))\
            .aggregate(max_len=Max(Length("code")))["max_len"]
        if cur_max_len is None:
            return
        new_max_len = cur_max_len - len(self.account.code)\
            + len(self.data["code"])
        if new_max_len <= 5:
            return
        error = forms.ValidationError(
            _("The descendant account codes will be too long  (max. 5)."),
            code="descendant_code_size")
        self.add_error("code", error)
        raise error



# Generated by Django 3.1 on 2020-08-27 15:07

import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('code', models.CharField(max_length=5, unique=True)),
                ('title_l10n', models.CharField(db_column='title', max_length=32)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_accounting_account', to=settings.AUTH_USER_MODEL)),
                ('parent', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='child_set', to='accounting.account')),
                ('updated_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='updated_accounting_account', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('date', models.DateField()),
                ('ord', models.PositiveSmallIntegerField(default=1)),
                ('notes', models.CharField(blank=True, max_length=128, null=True)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_accounting_transaction', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='updated_accounting_transaction', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('is_credit', models.BooleanField()),
                ('ord', models.PositiveSmallIntegerField()),
                ('summary', models.CharField(blank=True, max_length=128, null=True)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=18)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='accounting.account')),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_accounting_record', to=settings.AUTH_USER_MODEL)),
                ('transaction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accounting.transaction')),
                ('updated_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='updated_accounting_record', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='AccountL10n',
            fields=[
                ('id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=128)),
                ('language', models.CharField(max_length=7)),
                ('value', models.CharField(max_length=65535)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_accounting_accountl10n', to=settings.AUTH_USER_MODEL)),
                ('master', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='l10n_set', to='accounting.account')),
                ('updated_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='updated_accounting_accountl10n', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
    ]



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/6/29

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The data models of the Mia core application.

"""
from typing import Any, Dict, List

from dirtyfields import DirtyFieldsMixin
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from mia_core.utils import new_pk, Language


class RandomPkModel(models.Model):
    """The abstract data model that uses 9-digit random primary keys."""
    id = models.PositiveIntegerField(primary_key=True)

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if self.pk is None:
            self.pk = new_pk(self.__class__)
        super().save(force_insert=force_insert, force_update=force_update,
                     using=using, update_fields=update_fields)

    class Meta:
        abstract = True


class StampedModel(models.Model):
    """The abstract base model that has created_at, created_by, updated_at, and
    updated_by."""
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT,
        related_name="created_%(app_label)s_%(class)s")
    updated_at = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT,
        related_name="updated_%(app_label)s_%(class)s")

    def __init__(self, *args, **kwargs):
        self.current_user = None
        if "current_user" in kwargs:
            self.current_user = kwargs["current_user"]
            del kwargs["current_user"]
        super().__init__(*args, **kwargs)

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if self.current_user is None:
            raise AttributeError(
                F"Missing current_user in {self.__class__.__name__}")
        try:
            self.created_by
        except ObjectDoesNotExist as e:
            self.created_by = self.current_user
        self.updated_by = self.current_user
        super().save(force_insert=force_insert, force_update=force_update,
                     using=using, update_fields=update_fields)

    class Meta:
        abstract = True


class LocalizedModel(models.Model):
    """An abstract localized model."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if getattr(self, "_l10n", None) is None:
            self._l10n: Dict[str, Dict[str, Any]] = {}
        # Amends the is_dirty() method in DirtyFieldsMixin
        if isinstance(self, DirtyFieldsMixin):
            old_is_dirty = getattr(self, "is_dirty", None)

            def new_is_dirty(check_relationship=False, check_m2m=None) -> bool:
                """Returns whether the current data model is changed."""
                if old_is_dirty(check_relationship=check_relationship,
                                check_m2m=check_m2m):
                    return True
                default_language = self._get_default_language()
                for name in self._l10n:
                    for language in self._l10n[name]:
                        new_value = self._l10n[name][language]
                        if language == default_language:
                            if getattr(self, name + "_l10n") != new_value:
                                return True
                        else:
                            l10n_rec = self._get_l10n_set() \
                                .filter(name=name, language=language) \
                                .first()
                            if l10n_rec is None or l10n_rec.value != new_value:
                                return True
                return False
            self.is_dirty = new_is_dirty

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        """Saves the data model, along with the localized contents."""
        default_language = self._get_default_language()
        l10n_to_save: List[models.Model] = []
        if getattr(self, "_l10n", None) is not None:
            for name in self._l10n:
                for language in self._l10n[name]:
                    new_value = self._l10n[name][language]
                    if language == default_language:
                        setattr(self, name + "_l10n", new_value)
                    else:
                        current_value = getattr(self, name + "_l10n")
                        if current_value is None or current_value == "":
                            setattr(self, name + "_l10n", new_value)
                        l10n_rec = self._get_l10n_set()\
                            .filter(name=name, language=language)\
                            .first()
                        if l10n_rec is None:
                            l10n_to_save.append(self._get_l10n_set().model(
                                master=self, name=name,
                                language=language,
                                value=self._l10n[name][language]))
                        elif l10n_rec.value != new_value:
                            if getattr(self, name + "_l10n") == l10n_rec.value:
                                setattr(self, name + "_l10n", new_value)
                            l10n_rec.value = new_value
                            l10n_to_save.append(l10n_rec)
        super().save(force_insert=force_insert, force_update=force_update,
                     using=using, update_fields=update_fields)
        for l10n_rec in l10n_to_save:
            if isinstance(self, StampedModel)\
                    and isinstance(l10n_rec, StampedModel):
                l10n_rec.current_user = self.current_user
            l10n_rec.save(force_insert=force_insert, force_update=force_update,
                          using=using, update_fields=update_fields)

    def _get_l10n_set(self):
        """Returns the related localization data model."""
        l10n_set = getattr(self, "l10n_set", None)
        if l10n_set is None:
            raise AttributeError("Please define the localization data model.")
        return l10n_set

    def _get_default_language(self) -> str:
        """Returns the default language."""
        default = getattr(self.__class__, "DEFAULT_LANGUAGE", None)
        return Language.default().id if default is None else default

    def get_l10n(self, name: str) -> Any:
        """Returns the value of a localized field in the current language.

        Args:
            name: The field name.

        Returns:
            The value of this field in the current language.
        """
        return self.get_l10n_in(name, Language.current().id)

    def get_l10n_in(self, name: str, language: str) -> Any:
        """Returns the value of a localized field in a specific language.

        Args:
            name: The field name.
            language: The language ID.

        Returns:
            The value of this field in this language.
        """
        if getattr(self, "_l10n", None) is None:
            self._l10n: Dict[str, Dict[str, Any]] = {}
        if name not in self._l10n:
            self._l10n[name]: Dict[str, Any] = {}
        if language not in self._l10n[name]:
            if language != self._get_default_language():
                l10n_rec = self._get_l10n_set() \
                    .filter(name=name, language=language) \
                    .first()
                self._l10n[name][language] = getattr(self, name + "_l10n") \
                    if l10n_rec is None else l10n_rec.value
            else:
                self._l10n[name][language] = getattr(self, name + "_l10n")
        return self._l10n[name][language]

    def set_l10n(self, name: str, value: Any) -> None:
        """Sets the value of a localized field in the current language.

        Args:
            name: The field name.
            value: The value.
        """
        self.set_l10n_in(name, Language.current().id, value)

    def set_l10n_in(self, name: str, language: str, value: Any) -> None:
        """Sets the value of a localized field in a specific language.

        Args:
            name: The field name.
            language: The language ID.
            value: The value.
        """
        if getattr(self, "_l10n", None) is None:
            self._l10n: Dict[str, Dict[str, Any]] = {}
        if name not in self._l10n:
            self._l10n[name]: Dict[str, Any] = {}
        self._l10n[name][language] = value

    class Meta:
        abstract = True


class L10nModel(models.Model):
    """The abstract base localization model."""
    name = models.CharField(max_length=128)
    language = models.CharField(max_length=7)
    value = models.CharField(max_length=65535)

    class Meta:
        abstract = True



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/6/30

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The route settings of the accounting application.

"""

from django.urls import path, register_converter
from django.views.decorators.http import require_GET

from mia_core.views import RedirectView
from . import converters, views

register_converter(converters.PeriodConverter, "period")
register_converter(converters.AccountConverter, "account")
register_converter(converters.CashAccountConverter, "cash-account")
register_converter(converters.LedgerAccountConverter, "ledger-account")
register_converter(converters.TransactionTypeConverter, "txn-type")
register_converter(converters.TransactionConverter, "txn")
register_converter(converters.DateConverter, "date")

app_name = "accounting"
urlpatterns = [
    path("", require_GET(RedirectView.as_view(
        query_string=True,
        pattern_name="accounting:cash.home",
    )), name="home"),
    path("cash",
         views.CashDefaultView.as_view(), name="cash.home"),
    path("cash/<cash-account:account>/<period:period>",
         views.cash, name="cash"),
    path("cash-summary",
         views.CashSummaryDefaultView.as_view(), name="cash-summary.home"),
    path("cash-summary/<cash-account:account>",
         views.cash_summary, name="cash-summary"),
    path("ledger",
         views.LedgerDefaultView.as_view(), name="ledger.home"),
    path("ledger/<ledger-account:account>/<period:period>",
         views.ledger, name="ledger"),
    path("ledger-summary",
         views.LedgerSummaryDefaultView.as_view(), name="ledger-summary.home"),
    path("ledger-summary/<ledger-account:account>",
         views.ledger_summary, name="ledger-summary"),
    path("journal",
         views.JournalDefaultView.as_view(), name="journal.home"),
    path("journal/<period:period>",
         views.journal, name="journal"),
    path("trial-balance",
         views.TrialBalanceDefaultView.as_view(), name="trial-balance.home"),
    path("trial-balance/<period:period>",
         views.trial_balance, name="trial-balance"),
    path("income-statement",
         views.IncomeStatementDefaultView.as_view(),
         name="income-statement.home"),
    path("income-statement/<period:period>",
         views.income_statement, name="income-statement"),
    path("balance-sheet",
         views.BalanceSheetDefaultView.as_view(), name="balance-sheet.home"),
    path("balance-sheet/<period:period>",
         views.balance_sheet, name="balance-sheet"),
    path("search",
         views.SearchListView.as_view(), name="search"),
    path("transactions/<txn-type:txn_type>/create",
         views.TransactionFormView.as_view(), name="transactions.create"),
    path("transactions/<txn-type:txn_type>/<txn:txn>",
         views.TransactionView.as_view(), name="transactions.detail"),
    path("transactions/<txn-type:txn_type>/<txn:txn>/update",
         views.TransactionFormView.as_view(), name="transactions.update"),
    path("transactions/<txn:txn>/delete",
         views.TransactionDeleteView.as_view(), name="transactions.delete"),
    path("transactions/sort/<date:date>",
         views.TransactionSortFormView.as_view(), name="transactions.sort"),
    path("accounts",
         views.AccountListView.as_view(), name="accounts"),
    path("accounts/create",
         views.AccountFormView.as_view(), name="accounts.create"),
    path("accounts/<account:account>",
         views.AccountView.as_view(), name="accounts.detail"),
    path("accounts/<account:account>/update",
         views.AccountFormView.as_view(), name="accounts.update"),
    path("accounts/<account:account>/delete",
         views.account_delete, name="accounts.delete"),
    path("api/accounts",
         views.api_account_list, name="api.accounts"),
    path("api/accounts/options",
         views.api_account_options, name="api.accounts.options"),
]



# The setup.py
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/9/7

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mia-accounting",
    version="0.0.2",
    author="imacat",
    author_email="imacat@mail.imacat.idv.tw",
    description="A Django accounting application",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/imacat/mia-accounting",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Framework :: Django :: 3.0",
        "Topic :: Office/Business :: Financial :: Accounting",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["django", "django-dirtyfields", "titlecase",
                      "django-decorator-include"],
)
